--
-- ManureWhisk
--
-- @author	nilBrain
-- @date	19.05.2021
-- @node	Expandet ManureSystem from @Wopster
--
-- Copyright (c) nilBrain, 2021


ManureWhisk = {};
ManureWhisk.MOD_NAME = g_currentModName;

ManureWhisk.RAYCAST_MASK = 32 + 64 + 128 + 256 + 4096 + 8194;
ManureWhisk.RAYCAST_DISTANCE = 1;

function ManureWhisk.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(TurnOnVehicle, specializations)-- and g_manureSystem ~= nil;
end;

function ManureWhisk.registerFunctions(vehicleType)
    SpecializationUtil.registerFunction(vehicleType, "whiskRaycastCallback", ManureWhisk.whiskRaycastCallback);
    SpecializationUtil.registerFunction(vehicleType, "setMixingState", ManureWhisk.setMixingState);
end;

function ManureWhisk.registerEventListeners(vehicleType)
    SpecializationUtil.registerEventListener(vehicleType, "onLoad", ManureWhisk);
    SpecializationUtil.registerEventListener(vehicleType, "onDelete", ManureWhisk);
    SpecializationUtil.registerEventListener(vehicleType, "onUpdateTick", ManureWhisk);
end

function ManureWhisk.registerOverwrittenFunctions(vehicleType)
    SpecializationUtil.registerOverwrittenFunction(vehicleType, "getPowerMultiplier", ManureWhisk.getPowerMultiplier);
end;

function ManureWhisk:onLoad(savegame)
    self.spec_ManureWhisk = self[("spec_%s.manureWhisk"):format(ManureWhisk.MOD_NAME)];
    local spec = self.spec_ManureWhisk;
    local baseKey = "vehicle.manureWhisk";

    spec.isRaycastAllowed = true;
    spec.lastRaycastDistance = 0;
    spec.lastRaycastObject = nil;
    spec.impact = 0;
    spec.node = FS19_manureSystem.ManureSystemXMLUtil.getOrCreateNode(self, self.xmlFile, baseKey);
    spec.mixPerSecond = Utils.getNoNil(getXMLFloat(self.xmlFile, baseKey .. "#mixPerSecond"), 500) * 2
    spec.yOffset = Utils.getNoNil(getXMLFloat(self.xmlFile, baseKey .. "#yOffset"), 0);
    spec.rayCastDistance = Utils.getNoNil(getXMLFloat(self.xmlFile, baseKey .. "#rayCastDistance"), ManureWhisk.RAYCAST_DISTANCE);

    spec.effects = {};
	spec.effects = g_effectManager:loadEffect(self.xmlFile, "vehicle.manureWhisk.whiskEffects.effect", self.components, self, self.i3dMappings);
end;

function ManureWhisk:onDelete()
	local spec = self.spec_ManureWhisk;
	if self.isClient then
		g_effectManager:deleteEffects(spec.effects);
	end;
end;

function ManureWhisk:onUpdateTick(dt, isActiveForInput, isActiveForInputIgnoreSelection, isSelected)
    local spec = self.spec_ManureWhisk;
    if self.isClient and spec.node ~= nil then
        if spec.isRaycastAllowed then
            local x, y, z = getWorldTranslation(spec.node);
            local dx, dy, dz = localDirectionToWorld(spec.node, 0, 0, -1);

            if self:getIsTurnedOn() then
                if spec.lastRaycastObject == nil then
                    raycastAll(x, y, z, dx, dy, dz, "whiskRaycastCallback", spec.rayCastDistance, self, ManureWhisk.RAYCAST_MASK, true);
                end;
            else
				g_effectManager:stopEffects(spec.effects);
                spec.lastRaycastDistance = 0
                spec.lastRaycastObject = nil
            end;

			local r, g, b = 1, 0, 0;

			local object = spec.lastRaycastObject;
			if object ~= nil then
				r, g = 0, 1;
				if object:isUnderFillPlane(x, y + spec.yOffset, z) and self:getIsTurnedOn() then
					object:decreaseManureThickness(spec.mixPerSecond, dt, true, true);
					self:setMixingState(object.fillPlane.planeNode, spec.mixPerSecond * 0.5, object.thickness);
					g_effectManager:setFillType(spec.effects, Utils.getNoNil(FillType.LIQUIDMANURE, FillType.UNKNOWN));
					g_effectManager:startEffects(spec.effects);
				else
					spec.impact = 0;
					object.fillPlaneIsIdle = false;
					g_effectManager:stopEffects(spec.effects);
				end;


				if g_manureSystem.debug then
					local lx, ly, lz = worldToLocal(spec.node, x, y, z);
					lz = lz - ManureWhisk.RAYCAST_DISTANCE;
					lx, ly, lz = localToWorld(spec.node, lx, ly, lz);
					drawDebugLine(x, y, z, r, g, b, lx, ly, lz, r, g, b);
				end;
			end;
        end;
    end;
end;

function ManureWhisk:getPowerMultiplier(superFunc)
    local spec = self.spec_ManureWhisk;
    local object = spec.lastRaycastObject;
    if object ~= nil then
        local thickness = object.thickness;
        local mq = (1.1 - self:getFillUnitFillLevelPercentage()) * 0.001;
        return (thickness * mq) * 1000;
    end;
    return superFunc(self);
end

function ManureWhisk:whiskRaycastCallback(hitObjectId, x, y, z, distance)
    if hitObjectId ~= 0 then

        if hitObjectId == g_currentMission.terrainRootNode then
            return true
        end

        local object = g_currentMission:getNodeObject(hitObjectId);
        if object ~= nil and object.isa ~= nil then
            local spec = self.spec_ManureWhisk;

            if object:isa(Placeable) and object.isUnderFillPlane ~= nil then
                spec.lastRaycastDistance = distance;
                spec.lastRaycastObject = object;
                return false;
            end;
        end;
    end;

    return true;
end;

function ManureWhisk:setMixingState(planeNode, literPerSecond, thickness)
    local spec = self.spec_ManureWhisk;
    local _,_,z = getShaderParameter(planeNode, "displacementScaleSpeedFrequency");
    local impact = (literPerSecond * (1.1 - thickness) / literPerSecond)/1000;
    impact = MathUtil.clamp(z + impact, 0, 1);

    spec.impact = impact;
    local maxOffsetScale = (literPerSecond * 0.8) / 1000
    local offsetScale = maxOffsetScale * impact
    local waveY = impact * 1.1
    local waveZ = impact
    setShaderParameter(planeNode, "displacementScaleSpeedFrequency", offsetScale, waveY, waveZ, 0, false);
end;